package com.dx.union;

import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.io.resource.ResourceUtil;
import cn.hutool.core.text.csv.CsvReader;
import cn.hutool.core.text.csv.CsvUtil;
import cn.hutool.core.text.csv.CsvWriter;
import cn.hutool.core.thread.ThreadUtil;
import cn.hutool.core.util.NumberUtil;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.http.HttpRequest;
import cn.hutool.http.HttpResponse;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.File;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.util.List;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

/**
 *
 */
public class CheckLogin {

    private static final Logger log = LogManager.getLogger(CheckLogin.class);
    private static final char[] HEX_CHARS = new char[]{'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};
    private static final String DEFAULT_FILE_PATH = "/tmp/users.csv";
    /**
     * 1：立即执行
     * 0：间隔运行
     */
    private static final Integer DEFAULT_IS_RUN = 0;

    public static void main(String[] args) {
        try {
            log.info("开始运行打卡任务");
            String morning = DateUtil.today() + " 09:00:00";
            String afternoon = DateUtil.today() + " 18:00:00";
            DateTime morningDate = DateUtil.parse(morning);
            DateTime afternoonDate = DateUtil.parse(afternoon);
            DateTime now = DateUtil.date();
            DateTime beginOfDay = DateUtil.beginOfDay(now);
            DateTime endOfDay = DateUtil.endOfDay(now);
            log.info("当前时间:{}", now.toString());
            boolean isMorning = DateUtil.isIn(now, beginOfDay, morningDate);
            boolean isAfterNoon = DateUtil.isIn(now, afternoonDate, endOfDay);
            if (!isMorning && !isAfterNoon) {
                log.info("现在不是打卡时间哦！");
                return;
            }
            int singMark = isMorning ? 0 : 1;
            log.info("现在要打 {} 卡", isMorning ? "上班" : "下班");
            String filePath = args.length > 0 ? args[0] : DEFAULT_FILE_PATH;
            Integer isNowRun = args.length > 0 ? Integer.valueOf(args[1]) : DEFAULT_IS_RUN;
            if (!FileUtil.exist(filePath)) {
                log.error("未自定义账号文件路径（运行命令行添加自定义账号文件路径 例如：java -jar xxx.jar /use/local/user.csv），默认文件路径 {} 读取为空", DEFAULT_FILE_PATH);
                return;
            }
            final CsvReader reader = CsvUtil.getReader();
            final List<User> userList = reader.read(
                    ResourceUtil.getUtf8Reader(filePath), User.class);
            int size = userList.size();
            log.info("共有 {} 个需打卡账号", size);
            Set<User> userSet = RandomUtil.randomEleSet(userList, size);
            userSet.forEach(forEachWithIndex((user, index) -> {
                boolean todaySign = user.isTodaySign(singMark);
                if (todaySign) {
                    log.info("{} 账号已完成打卡", user.getUsername());
                    return;
                }
                boolean isSignSuccess = sign(user, singMark);
                if (isSignSuccess) {
                    log.info("{} 账号打卡成功", user.getUsername());
                    String sign_time = DateUtil.date().toString();
                    if (isMorning) {
                        user.setIsStartWorkSign(sign_time);
                    } else {
                        if (StrUtil.isEmpty(user.getIsStartWorkSign())) {
                            user.setIsStartWorkSign("");
                        }
                        user.setIsEndWorkSign(sign_time);
                    }
                } else {
                    log.error("{} 账号打卡失败", user.getUsername());
                }
                if (index + 1 == size) {
                    return;
                }
                // 随机 1-3分钟之后执行下一个人的打卡
                // 1 * 60 * 1000
                BigDecimal randomMinute = NumberUtil.round(RandomUtil.randomFloat(0.1F, 1.2F),2);
                if (isNowRun == 1) {
                    randomMinute = BigDecimal.valueOf(0);
                }
                log.info("{}分钟后 执行下个人的打卡任务", randomMinute);
                ThreadUtil.safeSleep(NumberUtil.mul(randomMinute, 60, 1000));
            }));
            CsvWriter writer = CsvUtil.getWriter(new File(filePath), StandardCharsets.UTF_8);
            writer.writeBeans(userList);
        } catch (Exception e) {
            log.error("运行出错！", e);
        } finally {
            log.info("打卡任务结束");
        }
    }

    public static boolean sign(User user, Integer signMark) {
        boolean result = false;
        try {
            HttpRequest.getCookieManager().getCookieStore().removeAll();
            String username = user.getUsername();
            String password = user.getPassword();
            log.info("开始{}打卡 {} 的账号，密码为：{}", signMark == 0 ? "上班" : "下班", username, password);
            password = encrypt2ToMD5(password);
            log.info("加密密码为：{}", password);
            try (HttpResponse response = HttpRequest.post("https://login.oa.unionpay.com/idsapi/portal/Login")
                         .form("username", username)
                         .form("password", password)
                         .form("saveUsername", "false")
                         .form("savePassword", "false")
                         .form("loginType", "normal")
                         .form("code", "")
                         .execute()) {
                JSONObject data = JSONUtil.parseObj(response.body()).getJSONObject("data");
                if (data != null) {
                    String token = data.getStr("token");
                    log.info("认证系统返回token {}", token);
                    try (HttpResponse thirdResponse = HttpRequest
                            .get("http://oms.oa.unionpay.com/prod-api/checkTokenThenLogin?personType=0&token=" + token)
                            .execute()) {
                        JSONObject thirdBodyJson = JSONUtil.parseObj(thirdResponse.body());
                        if (!"200".equals(thirdBodyJson.getStr("code"))) {
                            log.error("签到系统返回报文出错 {}", thirdBodyJson);
                            return false;
                        }
                        String thirdToken = thirdBodyJson.getStr("localToken");
                        try (HttpResponse httpResponse = HttpRequest.post("http://oms.oa.unionpay.com/prod-api/atds/ds/" + signMark)
                                .header("Authorization", "Bearer " + thirdToken)
                                .execute()) {
                            JSONObject signJson = JSONUtil.parseObj(httpResponse.body());
                            if ("200".equals(signJson.getStr("code"))) {
                                String signDateTime = DateUtil.formatDateTime(signJson.getDate("data"));
                                result = true;
                                log.info("{} 打卡时间 {}", user.getUsername(), signDateTime);
                            } else {
                                log.error("签到接口出错 {}", signJson);
                            }
                        }
                    }
                } else {
                    log.error("认证系统返回报文出错 {}", response.body());
                }
            }
        } catch (Exception e) {
            log.error("sign error", e);
        }
        return result;
    }

    /**
     * MD5 加密
     * @param str 待加密字符串
     * @return MD5加密后字符串
     */
    public static String encrypt2ToMD5(String str) {
        // 加密后的16进制字符串
        String hexStr = "";
        try {
            // 此 MessageDigest 类为应用程序提供信息摘要算法的功能
            MessageDigest md5 = MessageDigest.getInstance("MD5");
            // 转换为MD5码
            byte[] digest = md5.digest(str.getBytes(StandardCharsets.UTF_8));
            hexStr = toHexString(digest);
        } catch (Exception e) {
            log.error("encrypt2ToMD5 error", e);
        }
        return hexStr;
    }

    public static String toHexString(byte[] input) {
        StringBuilder result = new StringBuilder();
        for (byte b : input) {
            result.append(HEX_CHARS[b >>> 4 & 15]);
            result.append(HEX_CHARS[b & 15]);
        }
        return result.toString();
    }

    /**
     * 利用BiConsumer实现foreach循环支持index
     *
     */
    public static <T> Consumer<T> forEachWithIndex(BiConsumer<T, Integer> biConsumer) {
        /*
        这里说明一下，我们每次传入forEach都是一个重新实例化的Consumer对象，在lambada表达式中我们无法对int进行++操作,
        我们模拟AtomicInteger对象，写个getAndIncrement方法，不能直接使用AtomicInteger哦
        */
        class IncrementInt{
            int i = 0;
            public int getAndIncrement(){
                return i++;
            }
        }
        IncrementInt incrementInt = new IncrementInt();
        return t -> biConsumer.accept(t, incrementInt.getAndIncrement());
    }

}
